class OandaError(Exception):
    """Oanda Error Exception Class"""
    def __init__(self, status_code='', resp_content=''):
        """OandaError Exception raised when server returns error.
        Args:
            status_code (str): Status code retrieved from the server.
            resp_content (dict): Response's body with more detailed info about
                                 the problem that has occurred.

        """
        if resp_content:
            message = "OANDA server returned [{}] status code".format(status_code)

            if "errorCode" in resp_content:
                message += ", and API returned [{}] error code".format(resp_content["errorCode"])

            if "errorMessage" in resp_content:
                message += ", with message: ({})".format(resp_content["errorMessage"])

        else:
            message = "OANDA server returned a internal server error"

        super().__init__(message)


class EnvironmentNotFound(Exception):
    pass


class AuthError(OandaError):
    pass


class NotFound(OandaError):
    pass


class ServerError(OandaError):
    pass



from .oanda_base import Core  # noqa
from .account import Account  # noqa
from .instruments import Instrument  # noqa
from .orders import Orders  # noqa
from .positions import Positions  # noqa
from .pricing import Pricing  # noqa
from .trades import Trades  # noqa
from .transactions import Transactions  # noqa


__all__ = [
    'Core', 'Account', 'Instrument', 'Orders', 'Positions', 'Pricing',
    'Trades', 'Transactions'
]



"""
Pricing endpoint
"""
from ..factories import ResponseFactory


class Pricing:
    """Class holding pricing functions
    Pricing
        Docs: http://developer.oanda.com/rest-live-v20/pricing-ep/
    """

    def __init__(self, api):
        self._api = api

    def get_pricing(self, account_id, instruments, **kwargs):
        """Get pricing information for a specified list of Instruments within
        an Account.
        Args:
            account_id str: Account Identifier [required]
            instruments str: List of Instruments to get pricing for. [required]
            since dateTime: Date/Time filter to apply to the response.
                           Only prices and home conversions (if requested) with a time
                           later than this filter
            includeUnitsAvailable bol: Flag that enables the inclusion of the
                                       unitsAvailable field in the returned Price objects
            includeHomeConversions bol: Flag that enables the inclusion of the homeConversions
                                        field in the returned response
        Returns:
            A dict with pricing information.

        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/pricing'.format(account_id)
        if not isinstance(instruments, list):
            instruments = [instruments]
        else:
            instruments = "%2C".join(instruments)

        params = {'instruments': instruments}

        qs_params = ('since', 'includeUnitsAvailable', 'includeHomeConversions')
        for qs in qs_params:
            if qs not in kwargs:
                continue
            # Add the ordered parameters
            params[qs] = kwargs.get(qs)

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetPricing')

    def get_pricing_stream(self, account_id, instruments, **kwargs):
        """Get a stream of Account Prices starting from when the request is made.
        This pricing stream does not include every single price created for the Account,
        but instead will provide at most 4 prices per second (every 250 milliseconds)
        for each instrument being requested.
        Args:
            account_id str: Account Identifier [required]
            instruments str: List of Instruments to get pricing for. [required]
            since dateTime: Date/Time filter to apply to the response.
                           Only prices and home conversions (if requested) with a time
                           later than this filter
            snapshot bol: Flag that enables/disables the sending of a pricing snapshot
                          when initially connecting to the stream. [default=True]
        Returns:
            A dict with pricing information.

        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{}/pricing/stream'.format(account_id)

        if not isinstance(instruments, list):
            instruments = [instruments]
        else:
            instruments = "%2C".join(instruments)
        params = {'instruments': instruments}

        qs_params = ('since', 'includeUnitsAvailable', 'includeHomeConversions')
        for qs in qs_params:
            if qs not in kwargs:
                continue
            # Add the ordered parameters
            params[qs] = kwargs.get(qs)

        return self._api.search(endpoint, params=params, stream=True)



"""
Account endpoints
"""
from ..factories import ResponseFactory


class Account:
    """Class holding account functions.
    Account
        Docs: http://developer.oanda.com/rest-live-v20/account-ep/
    """
    def __init__(self, api):
        self._api = api

    def get_accounts(self):
        """Get a list of all Accounts authorized for the provided token.

        Args:
            This function takes no arguments.

        Returns:
            dict: All accounts ids and tags.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            OandaError: An error occurred while requesting the Oanda API.

        """

        endpoint = 'accounts'
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetAccounts')

    def get_account(self, account_id):
        """Get the full details for a single Account that a client has access
        to.

        Full pending Order, open Trade and open Position representations
        are provided.

        Args:
            account_id (str): Account ID.

        Returns:
            dict: Full account details.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            OandaError: An error occurred while requesting the Oanda API.

        """
        endpoint = 'accounts/{}'.format(account_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetAccount')

    def get_account_summary(self, account_id):
        """Get a summary for a single Account that a client has access to.

        Args:
            account_id: A string providing an account ID.

        Returns:
            A dict with an account summary.

        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/summary'.format(account_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetAccountSummary')

    def get_instruments(self, account_id, instruments=None):
        """Get the list of tradeable instruments for the given Account.
        Args:
            account_id: A string providing an account ID.
            instruments: An optional list of instruments to query specifically.
        Returns:
            A dict with a list of tradeable instruments for the account ID that
            has been provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/instruments'.format(account_id)
        params = {}

        if instruments:
            inst = "%2C".join(instruments)
            params["instruments"] = inst
        response = self._api.request(endpoint, params=params)
        return ResponseFactory(response, 'GetAccountInstruments')

    def set_account_settings(self, account_id, alias=None, margin_rate=None):
        """Set the client-configurable portions of an Account.
        Args:
            account_id: A string providing an account ID.
            alias: An optional string defining the name for the account.
            margin_rate: An optional string defining the margin rate for the
            account.
        Returns:
            A dict with operation response.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/configuration'.format(account_id)
        params = {}
        if alias:
            params["alias"] = alias

        if margin_rate:
            params["marginRate"] = margin_rate

        response = self._api.request(endpoint, "PATCH", params=params)
        return ResponseFactory(response, 'GetAccountSettings')

    def get_account_changes(self, account_id, since_transaction_id):
        """Endpoint used to poll an Account for its current state and changes
        since a specified TransactionID.
        Args:
            account_id: A string providing an account ID.
            since_transaction_id: A string providing an account ID.
        Returns:
            A dict with the account state and changes.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/changes'.format(account_id)
        params = {"sinceTransactionID": since_transaction_id}
        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetAccountChanges')



"""
Positions endpoints
"""
from ..factories import ResponseFactory


class Positions:
    """Class holding positions functions
    Positions
        Docs: http://developer.oanda.com/rest-live-v20/positions-ep/
    """
    def __init__(self, api):
        self._api = api

    def get_positions(self, account_id):
        """Get a List all Positions for an Account.
        The Positions returned are for every instrument
        that has had a position during the lifetime of an the Account.
        Args:
            account_id str: Account Identifier [required]
        Returns:
            A response factory with the Account’s Positions are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/positions'.format(account_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetPositions')

    def get_open_positions(self, account_id):
        """Get a List List all open Positions for an Account.
        An open Position is a Position in an Account that
        currently has a Trade opened for it..
        Args:
            account_id str: Account Identifier [required]
        Returns:
            A response factory with the Account’s Positions are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/openPositions'.format(account_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetOpenPositions')

    def get_position_details(self, account_id, instrument):
        """Get Get the details of a single Instrument’s Position in an Account.
        The Position may by open or not.
        Args:
            account_id str: Account Identifier [required]
            instrument str: Name of the Instrument [required
        Returns:
            A response factory with the Account’s Positions are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/positions/{1}'.format(account_id, instrument)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetPositionDetail')

    def close_position(self, account_id, instrument, long_units,
                       long_client_extensions, short_units,
                       short_client_extensions):
        """Closeout the open Position for a specific instrument in an Account.
        Args:
            This function takes no arguments.
        Returns:
            A response factory with the Account’s Positions are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/positions/{1}/close'.format(account_id, instrument)

        data = {}
        if long_units:
            data["longUnits"] = long_units

        if short_client_extensions:
            data["longClientExtensions"] = long_client_extensions

        if short_units:
            data["shortUnits"] = short_units

        if long_client_extensions:
            data["longClientExtensions"] = long_client_extensions

        response = self._api.update(endpoint, data=data)
        return ResponseFactory(response, 'GetClosePosition')



import status

from .exceptions import AuthError, NotFound, OandaError, ServerError


def validate_status_code(status_code, content):
    exception_class = None
    if status_code in (status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN):
        exception_class = AuthError(status_code, {'errorMessage': 'Authentication Error'})
    elif status_code == status.HTTP_404_NOT_FOUND:
        exception_class = NotFound(status_code, {'errorMessage': 'Endpoint Not Found'})
    elif status_code in range(status.HTTP_400_BAD_REQUEST, status.HTTP_417_EXPECTATION_FAILED + 1):
        exception_class = OandaError(status_code, content)
    elif status.is_server_error(status_code):
        exception_class = ServerError(status_code)

    if exception_class:
        return False, exception_class
    return True, exception_class



"""
Instruments endpoint
"""
from ..factories import ResponseFactory

API_DATE_ARGS = {
    'from_date': 'from',
    'to_date': 'to',
}


class Instrument:
    """Class holding instruments functions
    Instrument
        Docs: http://developer.oanda.com/rest-live-v20/instrument-ep/
    """

    def __init__(self, api):
        self._api = api

    def get_candles(self, instrument, **kwargs):
        """Fetch a Candlestick data for a instrument.

        Args:
            instrument:	Name of the Instrument [required]

        optional:
            price:	The Price component(s) to get candlestick data for.
                    Can contain any combination of the characters “M” (midpoint candles)
                    “B” (bid candles) and “A” (ask candles). [default=M]
            granularity: The granularity of the candlesticks to fetch [default=S5]
            count: The number of candlesticks to return in the response.
                  Count should not be specified if both the start and end parameters are
                  provided, as the time range combined with the granularity will determine
                  the number of candlesticks to return. [default=500, maximum=5000]
            from_date: The start of the time range to fetch candlesticks for.
            to_date: The end of the time range to fetch candlesticks for.
            smooth:	A flag that controls whether the candlestick is “smoothed” or not.
                    A smoothed candlestick uses the previous candle’s close price as its open price,
                    while an unsmoothed candlestick uses the first price from its
                    time range as its open price. [default=False]
            includeFirst: A flag that controls whether the candlestick that is covered by
                         the from time should be included in the results.
                         This flag enables clients to use the timestamp of the last
                         completed candlestick received to poll for future candlesticks
                         but avoid receiving the previous candlestick repeatedly. [default=True]
            dailyAlignment: The hour of the day (in the specified timezone) to use
                            for granularities that have daily alignments. [default=17, minimum=0, maximum=23]
            alignmentTimezone: The timezone to use for the dailyAlignment parameter.
                              Candlesticks with daily alignment will be aligned to the dailyAlignment
                              hour within the alignmentTimezone. Note that the returned times will
                              still be represented in UTC. [default=America/New_York]
            weeklyAlignment: The day of the week used for granularities that have weekly alignment. [default=Friday]


        Returns:
            OANDAFactoryResponse: A object with instrument information.

        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'instruments/{0}/candles'.format(instrument)
        qs_params = (
            'price', 'granularity', 'count', 'to_date', 'from_date',
            'smooth', 'includeFirst', 'dailyAlignment', 'alignmentTimezone',
            'weeklyAlignment',
        )
        params = {}
        for qs in qs_params:
            if qs not in kwargs:
                continue

            # Change the dates params
            param = API_DATE_ARGS.get(qs, qs)
            # Add the ordered parameters
            params[param] = kwargs.get(qs)

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetInstrumentCandles')

    def get_orderbook(self, instrument, **kwargs):
        """Fetch a OrderBook for an instrument.
        Args:
            instrument:	Name of the Instrument [required]
        optional:
            time: The time of the snapshot to fetch.
                  If not specified, then the most recent snapshot is fetched.
        Returns:
            OANDAFactoryResponse: A object with instrument information.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'instruments/{0}/orderBook'.format(instrument)

        params = {}
        if 'time' in kwargs:
            params['time'] = kwargs.get('time')

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetInstrumentOrderBook')

    def get_positionbook(self, instrument, **kwargs):
        """Fetch a position book for an instrument.
        Args:
            instrument:	Name of the Instrument [required]
        optional:
            time: The time of the snapshot to fetch.
                  If not specified, then the most recent snapshot is fetched.
        Returns:
            OANDAFactoryResponse: A object with instrument information.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'instruments/{0}/positionBook'.format(instrument)

        params = {}
        if 'time' in kwargs:
            params['time'] = kwargs.get('time')

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetInstrumentPositionbook')



from simple_model.builder import model_builder


class ResponseFactory:
    """"
    Return a OANDA Response object
    """
    def __init__(self, response, response_type=None):
        self.response_type = response_type or 'OandaResult'
        self.response = response

    def __repr__(self):
        return '<{} object>'.format(self.response_type)

    def as_dict(self):
        return self.response

    def as_obj(self):
        result = self.as_dict()
        return model_builder(result)



"""
Transactions endpoint
"""
from ..factories import ResponseFactory


API_PARAMS = {
    'from_date': 'from',
    'to_date': 'to',
    'page_size': 'pageSize',
    'type_list': 'type',
}


class Transactions:
    """Class holding transactions functions

    Transactions
        Docs: http://developer.oanda.com/rest-live-v20/transactions-ep/
    """

    def __init__(self, api):
        self._api = api

    def get_transactions(self, account_id, **kwargs):
        """Get a list of Transactions pages that satisfy a time-based Transaction query.
        Args:
            account_id (str): Account Identifier [required]
        Kwargs:
            from_date (DateTime): The starting time (inclusive) of the time range
                                  for the Transactions being queried. [default=Account Creation Time]
            to_date (DateTime): The ending time (inclusive) of the time range for the
                                Transactions being queried. [default=Request Time]
            page_size (int): The number of Transactions to include in each page of the results.
                            [default=100, maximum=1000]
            type_list (list): A filter for restricting the types of Transactions to retrieve.

        Returns:
            OANDAResponseFactory with the requested time range of Transaction pages are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/transactions'.format(account_id)

        params = {}
        qs_params = ('from_date', 'to_date', 'page_size', 'type_list')
        for qs in qs_params:
            if qs not in kwargs:
                continue

            param = API_PARAMS.get(qs, qs)

            if qs == 'type_list':
                if not isinstance(qs, list):
                    qs = [qs]
                else:
                    qs = "%2C".join(qs)

            # Add the ordered parameters
            params[param] = kwargs.get(qs)

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetTransactions')

    def get_transaction_details(self, account_id, transaction_id):
        """Get the details of a single Account Transaction.
        Args:
            account_id (str):	Account Identifier [required]
            transaction_id (str): A Transaction ID [required]
        Returns:
            OANDAResponseFactory with the details of the requested Transaction are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/transactions{1}'.format(account_id,
                                                         transaction_id)

        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetTransactionDetail')

    def get_transaction_idrange(self, account_id, **kwargs):
        """Get a range of Transactions for an Account based on the Transaction IDs.
        Args:
            account_id (str): Account Identifier [required]
        Kwargs:
            from_date (DateTime): The starting time (inclusive) of the time range
                                  for the Transactions being queried. [default=Account Creation Time]
            to_date (DateTime): The ending time (inclusive) of the time range for the
                                Transactions being queried. [default=Request Time]
            type_list (list): A filter for restricting the types of Transactions to retrieve.


        Returns:
            OANDAResponseFactory with the requested time range of Transactions are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/transactions/idrange'.format(account_id)

        params = {}
        qs_params = ('from_date', 'to_date', 'type_list')
        for qs in qs_params:
            if qs not in kwargs:
                continue

            param = API_PARAMS.get(qs, qs)

            if qs == 'type_list':
                if not isinstance(qs, list):
                    qs = [qs]
                else:
                    qs = "%2C".join(qs)

            # Add the ordered parameters
            params[param] = kwargs.get(qs)

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetTransactionIdRange')

    def get_transaction_sinceid(self, account_id, last_transaction_id):
        """Get a range of Transactions for an Account starting at (but not
        including) a provided Transaction ID.
        Args:
            account_id (str): Account Identifier [required]
            last_transaction_id (str): The ID of the last Transaction fetched.
                                       This query will return all Transactions newer
                                       than the TransactionID. [required]
        Returns:
            OANDAResponseFactory with the requested time range of Transactions are provided.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/transactions/sinceid'.format(account_id)

        params = {"id": last_transaction_id}
        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetTransactionSinceId')

    def get_transaction_stream(self, account_id):
        """Get a stream of Transactions for an Account starting from when the request is made.
        Args:
            account_id (str): Account Identifier [required]
        Returns:
            Connecting to the Transaction Stream was successful.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/transactions/stream'.format(account_id)
        return self._api.search(endpoint, stream=True)



"""
Orders endpoint
"""
from ..factories import ResponseFactory


class Orders:
    """Class holding orders functions
    Orders
        Docs: http://developer.oanda.com/rest-live-v20/orders-ep/
    """

    def __init__(self, api):
        self._api = api

    def create_order(self, account_id, order):
        """Get a list of all Accounts authorized for the provided token.
        Create an order for an Account

        Args:
            account_id str: Account Identifier [required]
            order dict: Order specification to create a order

        Returns:
            OANDAFactoryResponse: A object with instrument information.OANDAResponseFactory with order data.

        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/orders'.format(account_id)
        data = {'order': order}
        response = self._api.create(endpoint, data=data)
        return ResponseFactory(response, 'CreateOrderResponse')

    def get_orders_list(self, account_id, ids, state=None, instrument=None,
                        count=None, before_id=None):
        """Get a list of all Accounts authorized for the provided token.
        Get a list of Orders for an Account
        Args:
            This function takes no arguments.
        Returns:
            OANDAFactoryResponse: A object with instrument information.OANDAResponseFactory with order data.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/orders'.format(account_id)
        ids = "%2C".join(ids)
        params = {"ids": ids}

        if state:
            params["state"] = state

        if instrument:
            params["instrument"] = instrument

        if count:
            params["count"] = count

        if before_id:
            params["beforeID"] = before_id

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetOrdersList')

    def get_pending_orders(self, account_id):
        """Get a list of all Accounts authorized for the provided token.
        List all pending Orders in an Account
        Args:
            This function takes no arguments.
        Returns:
            OANDAFactoryResponse: A object with instrument information.OANDAResponseFactory with order data.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/pendingOrders'.format(account_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetPendingOrders')

    def get_order_details(self, account_id, order_id):
        """Get a list of all Accounts authorized for the provided token.
        Get details for a single Order in an Account
        Args:
            This function takes no arguments.
        Returns:
            OANDAResponseFactory with order data.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/orders/{1}'.format(account_id, order_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetOrderDetail')

    def replace_order(self, account_id, order_id, order):
        """Get a list of all Accounts authorized for the provided token.
        Replace an Order in an Account by simultaneously cancelling it and
        creating a replacement Order
        Args:
            This function takes no arguments.
        Returns:
            OANDAResponseFactory with order data.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/orders/{1}'.format(account_id, order_id)
        data = {'order': order}
        response = self._api.update(endpoint, data=data)
        return ResponseFactory(response, 'ReplaceOrder')

    def cancel_pending_order(self, account_id, order_id):
        """Get a list of all Accounts authorized for the provided token.
        Cancel a pending Order in an Account
        Args:
            This function takes no arguments.
        Returns:
            OANDAResponseFactory with order data.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{}/orders/{}/cancel'.format(account_id, order_id)
        response = self._api.update(endpoint, data={})
        return ResponseFactory(response, 'GetCancelPendingOrder')

    def update_client_extensions(self, account_id, order_id, comment):
        """Update the Client Extensions for an Order in an Account.
        Do not set, modify, or delete clientExtensions if your account is associated with MT4.
        Args:
            This function takes no arguments.
        Returns:
            OANDAResponseFactory with order data.
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{}/orders/{}/clientExtensions'.format(account_id, order_id)
        data = {
            "clientExtensions": {
                "comment": comment
            }
        }
        response = self._api.update(endpoint, data=data)
        return ResponseFactory(response, 'UpdateOrderClientExtensions')



"""
Trades endpoint
"""
from ..factories import ResponseFactory

API_PARAMS = {
    'before_id': 'beforeID',
    'take_profit': 'takeProfit',
    'stop_loss': 'stopLoss',
    'trailing_stoploss': 'trailingStopLoss'
}


class Trades:
    """Class holding trades functions
    Trades
        Docs: http://developer.oanda.com/rest-live-v20/trades-ep/
    """

    def __init__(self, api):
        self._api = api

    def get_trades(self, account_id, ids, **kwargs):
        """Get a list of all Accounts authorized for the provided token.
        Get a list of Trades for an Account

        Args:
            account_id (str): Account Identifier [required]
            ids (str, list): List of Trade IDs to retrieve.
        Kwargs:
            state: The state to filter the requested Trades by. [default=OPEN]
            instrument (str): The instrument to filter the requested Trades by.
            count (int): The maximum number of Trades to return. [default=50, maximum=500]
            before_id (str): The maximum Trade ID to return.
                            If not provided the most recent Trades in the Account are returned.
        Returns:
            OANDAResponseFactory with the list of Trades requested
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/trades'.format(account_id)
        if not isinstance(ids, list):
            ids = [ids]
        else:
            ids = "%2C".join(ids)

        params = {"ids": ids}
        qs_params = ('state', 'instrument', 'count', 'before_id')
        for qs in qs_params:
            if qs not in kwargs:
                continue

            param = API_PARAMS.get(qs, qs)
            # Add the ordered parameters
            params[param] = kwargs.get(qs)

        response = self._api.search(endpoint, params=params)
        return ResponseFactory(response, 'GetTrades')

    def get_open_trades(self, account_id):
        """Get the list of open Trades for an Account
        Args:
            account_id (str): Account Identifier [required]
        Returns:
            OANDAResponseFactory with the Account’s list of open Trades is provided
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/openTrades'.format(account_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetOpenTrades')

    def get_trade_details(self, account_id, trade_id):
        """Get the details of a specific Trade in an Account
        Args:
            account_id (str): Account Identifier [required]
            trade_id (str): Specifier for the Trade [required]
        Returns:
            OANDAResponseFactory with the details for the requested Trade is provided

        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/trades/{1}'.format(account_id, trade_id)
        response = self._api.search(endpoint)
        return ResponseFactory(response, 'GetTradeDetail')

    def close_trade(self, account_id, trade_id, units):
        """Close (partially or fully) a specific open Trade in an Account
        Args:
            account_id (str): Account Identifier [required]
            trade_id (str): Specifier for the Trade [required]
            units: Indication of how much of the Trade to close. ALL to close all.
        Returns:
            OANDAResponseFactory with the Trade has been closed as requested
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/trades/{1}/close'.format(account_id, trade_id)

        data = {"units": units}
        response = self._api.update(endpoint, data=data)
        return ResponseFactory(response, 'GetTradeDetail')

    def update_client_extensions(self, account_id, trade_id, client_extensions):
        """Update the Client Extensions for a Trade.
        Do not add, update, or delete the Client Extensions
        if your account is associated with MT4.

        Args:
            account_id (str): Account Identifier [required]
            trade_id (str): Specifier for the Trade [required]
            client_extensions (dict): The Client Extensions to update the Trade with.
        Returns:
            OANDAResponseFactory with the Trade has been closed as requested
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/trades/{1}/clientExtensions'.format(account_id,
                                                                     trade_id)
        data = {"clientExtensions": client_extensions}
        response = self._api.update(endpoint, data=data)
        return ResponseFactory(response, 'UpdateClientExtensions')

    def update_trade(self, account_id, trade_id, **kwargs):
        """Create, replace and cancel a Trade’s dependent Orders
        (Take Profit, Stop Loss and Trailing Stop Loss) through the Trade itself

        Args:
            account_id (str): Account Identifier [required]
            trade_id (str): Specifier for the Trade [required]
        Kwargs:
            take_profit (dict): The specification of the Take Profit to create/modify/cancel
            stop_loss (dict): The specification of the Stop Loss to create/modify/cancel.
            trailing_stop_loss (dict): The specification of the Trailing Stop Loss to create/modify/cancel
        Returns:
            OANDAResponseFactory with the Trade has been closed as requested
        Raises:
            OandaError: An error occurred while requesting the OANDA API.
        """
        endpoint = 'accounts/{0}/trades/{1}/orders'.format(account_id, trade_id)
        data = {}
        qs_params = ('take_profit', 'stop_loss', 'trailing_stop_loss')
        for qs in qs_params:
            if qs not in kwargs:
                continue

            param = API_PARAMS.get(qs, qs)
            # Add the ordered parameters
            data[param] = kwargs.get(qs)

        response = self._api.update(endpoint, data=data)
        return ResponseFactory(response, 'UpdateTrade')



import json

import requests
from requests.exceptions import Timeout, ReadTimeout, ConnectionError as RequestsConnectionError

from ..exceptions import EnvironmentNotFound, ServerError
from ..validators import validate_status_code


class RequestsTransport:
    def __init__(self, headers):
        self.session = requests.Session()
        self.session.headers.update(headers)

    def make_request(self, method_name, endpoint, **kwargs):
        method = getattr(self.session, method_name.lower())
        return method(endpoint, **kwargs)


class Core:
    """
    Core Abstract object.
    Attributes:
        url_bases: URLs from OANDA environments
        api_version: OANDA's API Version
        timeout: Request timeout
        token: OANDA's API Access token
        transport: Transport class to able to use Requests or AioHTTP(AsyncIO)
    """
    url_bases = {
        'live': 'https://api-fxtrade.oanda.com',
        'practice': 'https://api-fxpractice.oanda.com',
        'stream': 'https://stream-fxtrade.oanda.com'
    }
    api_version = 'v3'

    def __init__(self, environment, access_token, timeout=3, **kwargs):
        """
        Args:
            environment (str): Provides the environment for OANDA's API.
            access_token (str): Specifies the access token.
            timeout (int): Set the request timeout. Default is 3
        """
        self.base_url = self._get_base_url(environment)
        self.timeout = timeout
        self.token = access_token
        self.transport = RequestsTransport(headers=self.get_default_headers())

    def _get_base_url(self, environment):
        url_base = self.url_bases.get(environment)
        if not url_base:
            raise EnvironmentNotFound("Environment '{0}' does not exist!".format(environment))
        return "/".join((url_base, self.api_version))

    def get_default_headers(self):
        """Get the required headers to access the API
        Returns:
             dict: Authorization and Content Type
        """
        return {
            'Authorization': 'Bearer {}'.format(self.token),
            'Content-Type': 'application/json'
        }

    def make_request(self, method_name, endpoint, stream=False, **kwargs):
        """Requests data from Oanda API.
        Args:
            endpoint (str): URL for Oanda API endpoint.
            method_name (str): Specifies the method to be used on the request.
            stream (str): Set request as Streamer URL
        Optional:
            params (dict, optional): Specifies parameters to be sent with the
            request.

        Returns:
            response: Requests response object.
        """
        base_url = self.base_url
        if stream:
            base_url = self._get_base_url('stream')

        full_url = "/".join((base_url, endpoint))
        try:
            response = self.transport.make_request(
                method_name, full_url, timeout=self.timeout, stream=stream, **kwargs
            )
        except (Timeout, ReadTimeout, RequestsConnectionError) as exc:
            raise ServerError() from exc

        is_valid, exception_class = validate_status_code(
            response.status_code, response.json()
        )
        if not is_valid:
            raise exception_class

        return response

    def create(self, endpoint, data):
        """Do a POST without need to pass all arguments to make a request
        Args:
            endpoint (str): URL for Oanda API endpoint.
            data (dict, list of tuples): Data to send in the body of the request.

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        response = self.make_request('POST', endpoint, data=json.dumps(data))
        return response.json()

    def update(self, endpoint, data, partial=False):
        """Do a update (PUT/PATCH) without need to pass all arguments to make a request
        Args:
            endpoint (str): URL for Oanda API endpoint.
            data (dict, list of tuples): Data to send in the body of the request.
            partial (bool): To specify whether the update will change everything
                            or just a few attributes. Default is False

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        method = 'PATCH' if partial else 'PUT'
        response = self.make_request(method, endpoint, data=json.dumps(data))
        return response.json()

    def search(self, endpoint, stream=False, **kwargs):
        """Do a GET to make a search without need
           to pass all arguments to make a request.
        Args:
            endpoint (str): URL for Oanda API endpoint.
            stream (str): Set request as Streamer URL
        Kwargs(Optional):
            params (dict): Specifies parameters to be sent with the request.

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        params = kwargs.pop('params', {})
        response = self.make_request('GET', endpoint, params=params, stream=stream)
        return response.json()

    class Meta:
        abstract = True



"""Oanda API Library"""

from .api import (
    Core, Account, Instrument, Orders, Positions, Pricing, Trades, Transactions
)


class APIv20(Core):
    """Oanda REST API v20 Class

    This class instantiates all endpoint classes

    """
    def __init__(self, environment="practice", access_token=None, timeout=3):
        """APIv20 object to communicate with Oanda REST API.

        Args:
            environment (str, optional): Provides the environment for
                OANDA's API. Defaults to 'practice'.
            access_token (str): Specifies the access token.

        """
        super().__init__(environment, access_token, timeout)

        self.account = Account(self)
        self.orders = Orders(self)
        self.positions = Positions(self)
        self.pricing = Pricing(self)
        self.trades = Trades(self)
        self.transactions = Transactions(self)
        self.instrument = Instrument(self)
